/* lmpnodes.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include "lmp.h"
#include "lmpmath.h"

#include <stdio.h>

/* free node and its subnodes */
int lmp_node_free (lmp_node_t * p) {
  int i;

  if (p != NULL) {
    switch (p->what) {
    case OP:
      for (i = 0 ; i < p->data.operation.argc; i++)
	lmp_node_free (p->data.operation.argv[i]);
      break;
    case CONTROL:
      for (i = 0 ; i < p->data.control.argc; i++)
	lmp_node_free (p->data.control.argv[i]);
      break;
    case VAR:
      free (p->data.variable); // Hope this is a copy, not a pointer to something unique
      break;
    case VAL:
      break;
    }
    free (p);
  }

  return 0;
}

#ifndef _STDARG_FP_SAFE
lmp_node_t * lmp_node_new_fp_safe (lmp_node_type nt, lmp_value_type value) {
  lmp_node_t * node;

  node = malloc (sizeof (lmp_node_t));
  node->what = nt;

  node->data.value = value;

  return node;
}
#endif

lmp_node_t * lmp_node_new (lmp_node_type nt, ...) {
  lmp_node_t * node;
  char * s;
  int i;
  va_list args;

  va_start (args, nt);
  node = malloc (sizeof (lmp_node_t));
  node->what = nt;

  switch (nt) {
  case OP:
    node->data.operation.op   = va_arg (args, lmp_operator);
    node->data.operation.argc = va_arg (args, int);
    node->data.operation.argv = malloc (sizeof (lmp_node_t *) * node->data.operation.argc);
    i = 0;
    while (i < node->data.operation.argc)
      node->data.operation.argv [i++] = va_arg (args, lmp_node_t *);
    break;
  case CONTROL:
    node->data.control.op   = va_arg (args, lmp_control);
    node->data.control.argc = va_arg (args, int);
    node->data.control.argv = malloc (sizeof (lmp_node_t *) * node->data.operation.argc);
    i = 0;
    while (i < node->data.operation.argc)
      node->data.operation.argv [i++] = va_arg (args, lmp_node_t *);
    break;
  case VAL:
#if defined (_STDARG_FP_SAFE) || ! defined (HAVE_COMPLEX)
    node->data.value = va_arg (args, lmp_value_type);
#else if
    node->data.value =  va_arg (args, lmp_double_t);
    node->data.value += _Complex_I * (lmp_value_type) va_arg (args, lmp_double_t);
#endif

    break;
  case VAR:
    s = va_arg (args, char *);
    // copy id instead of linking, it's more save
    node->data.variable = malloc (strlen (s) + 1);
    strcpy (node->data.variable, s);
    break;
  }

  va_end (args);

  return node;
}

/* create a clone */
lmp_node_t * lmp_node_copy (lmp_node_t * n) {
  lmp_node_t * res;
  int i;

  res = malloc (sizeof (lmp_node_t));
  res->what = n->what;
  switch (res->what) {
  case OP:
    res->data.operation.op   = n->data.operation.op; 
    res->data.operation.argc = n->data.operation.argc;
    res->data.operation.argv = malloc (sizeof (lmp_node_t *) * n->data.operation.argc);
    /* copy subnodes */
    for (i = 0; i < n->data.operation.argc; i++) {
      res->data.operation.argv [i] = lmp_node_copy (n->data.operation.argv [i]);
    }
    break;
  case CONTROL:
    res->data.control.op   = n->data.control.op; 
    res->data.control.argc = n->data.control.argc;
    res->data.control.argv = malloc (sizeof (lmp_node_t *) * n->data.control.argc);
    for (i = 0; i < n->data.control.argc; i++) {
      res->data.control.argv [i] = lmp_node_copy (n->data.control.argv [i]);
    }
    break;
  case VAL:
    res->data.value = n->data.value;
    break;
  case VAR:
    /* even _copy_ variable id */
    res->data.variable = malloc (strlen (n->data.variable) + 1);
    strcpy (res->data.variable, n->data.variable);
    break;
  }

  return res;
}

int lmp_node_factored (lmp_node_t * node, lmp_node_t * var, lmp_node_t ** fact) {
  lmp_node_t * factor = NULL;
  switch (node->what) {
  case CONTROL:
    break;
  case OP:
    switch (node->data.operation.op) {
    case LMP_MULTIPLICATION:
      if (!lmp_tree_contains (node->data.operation.argv [0], var)) {
	if (lmp_node_factored (node->data.operation.argv [1], var, &factor)) {
	  *fact = lmp_node_new (OP, LMP_MULTIPLICATION, 2, factor, node->data.operation.argv [0]);
	  return 1;
	}
	else
	  lmp_node_free (factor);
      }
      if (!lmp_tree_contains (node->data.operation.argv [1], var)) {
	if (lmp_node_factored (node->data.operation.argv [0], var, &factor)) {
	  *fact = lmp_node_new (OP, LMP_MULTIPLICATION, 2, factor, node->data.operation.argv [1]);
	  return 1;
	}
	else
	  lmp_node_free (factor);
      }
      break;
    case LMP_DIVISION:
      if (!lmp_tree_contains (node->data.operation.argv [1], var)) {
	if (lmp_node_factored (node->data.operation.argv [0], var, &factor)) {
	  *fact = lmp_node_new (OP, LMP_DIVISION, 2, factor, node->data.operation.argv [1]);
	  return 1;
	}
	else
	  lmp_node_free (factor);
      }
      break;
    default:
      break;
    }
    break;
  case VAR:
    if (!strcmp (node->data.variable, var->data.variable)) {
      *fact = NEW_VAL (ONE);
      return 1;
    }
    break;
  case VAL:
    if (node->data.value == ZERO) {
      *fact = NEW_VAL (ZERO);
      return 1;
    }
    break;
  }

  return 0;
}

int lmp_node_polynom_fragment (lmp_node_t * node, lmp_node_t * var, lmp_node_t ** fact, lmp_node_t ** exp) {
  switch (node->what) {
  case OP:
    if (!lmp_tree_contains (node, var))
      break;
    switch (node->data.operation.op) {
    case LMP_MULTIPLICATION:
      if (lmp_tree_contains (node->data.operation.argv [0], var)) {
	if (lmp_tree_contains (node->data.operation.argv [1], var)) {
	  return 0;
	}
	else {
	  if (node->data.operation.argv [0]->what == OP) {
	    if (node->data.operation.argv [0]->data.operation.op == LMP_POWER) {
	      lmp_node_t
		* r = node->data.operation.argv [0]->data.operation.argv [1],
		* l = node->data.operation.argv [0]->data.operation.argv [0];

	      if (l->what == VAR && !lmp_tree_contains (r, var)) {
		if (!strcmp (l->data.variable, var->data.variable))
		  *exp = lmp_node_copy (r);
		else
		  *exp = lmp_node_copy (node->data.operation.argv [0]);

		*fact = lmp_node_copy (node->data.operation.argv [1]);
		return 1;
	      }
	      else if (!lmp_tree_contains (r, var) && lmp_tree_contains (l, var)) {
		*exp = NEW_VAL (ZERO);
		*fact = lmp_node_copy (node->data.operation.argv [0]);
		return 1;
	      }
	      else
		return 0;
	    }
	    else
	      return lmp_node_polynom_fragment (node->data.operation.argv [0], var, fact, exp);
	  }
	}
      }
      else {
	if (node->data.operation.argv [1]->what == OP) {
	  if (node->data.operation.argv [1]->data.operation.op == LMP_POWER) {
	    lmp_node_t
	      * r = node->data.operation.argv [1]->data.operation.argv [1],
	      * l = node->data.operation.argv [1]->data.operation.argv [0];

	    if (l->what == VAR && !lmp_tree_contains (r, var)) {
	      if (!strcmp (l->data.variable, var->data.variable))
		*exp = lmp_node_copy (r);
	      else
		*exp = lmp_node_copy (node->data.operation.argv [1]);

	      *fact = lmp_node_copy (node->data.operation.argv [0]);
	      return 1;
	    }
	    else if (!lmp_tree_contains (r, var) && lmp_tree_contains (l, var)) {
	      *exp = NEW_VAL (ZERO);
	      *fact = lmp_node_copy (node->data.operation.argv [1]);
	      return 1;
	    }
	    else
	      return 0;
	  }
	  else
	    return lmp_node_polynom_fragment (node->data.operation.argv [1], var, fact, exp);
	}
	else {
	  if (node->data.operation.argv [1]->what == VAR && !strcmp (node->data.operation.argv [1]->data.variable, var->data.variable)) {
	    *fact = lmp_node_copy (node->data.operation.argv [0]);
	    *exp  = NEW_VAL (ONE);
	  }
	  else {
	    *fact = NEW_VAL (ZERO);
	    *exp  = lmp_node_copy (node);
	  }

	  return 1;
	}
      }
      break;

    case LMP_POWER:
      if (node->data.operation.argv [0]->what == VAR && !lmp_tree_contains (node->data.operation.argv [1], var)) {
	// XXX:INCOMPLETE: add a ^ -1 catcher
	if (lmp_tree_evaluable (node->data.operation.argv [1]) && lmp_eval (node->data.operation.argv [1]) == (lmp_value_type) -1.0) {
	  return 0;
	}
	else {
	  *fact = NEW_VAL (ONE);
	  *exp  = lmp_node_copy (node->data.operation.argv [1]);
	}

	return 1;
      }
      else if (!lmp_tree_contains (node->data.operation.argv [0], var) && !lmp_tree_contains (node->data.operation.argv [1], var)) {
	*fact = lmp_node_copy (node);
	*exp  = NEW_VAL (ZERO);

	return 1;
      }
      else
	return 0;
      break;

    default:
      if (lmp_tree_contains (node, var))
	return 0;
      break;
    }

    break;
  case CONTROL:
    break;
  case VAR:
    if (!strcmp (node->data.variable, var->data.variable)) {
      *exp  = NEW_VAL (ONE);
      *fact = NEW_VAL (ONE);
      return 1;
    }
    break;
  case VAL:
    break;
  }

  *exp  = NEW_VAL (ZERO);
  *fact = lmp_node_copy (node);

  return 1;
}

int lmp_node_polynom (lmp_node_t *node, lmp_node_t *var, lmp_stack_t * fact, lmp_stack_t * exp) {
  lmp_node_t * expel, * factel;
  int res;

  if (node->what == OP) {
    switch (node->data.operation.op) {
    case LMP_ADDITION:
    case LMP_SUBTRACTION:
      return lmp_node_polynom (node->data.operation.argv [0], var, exp, fact) && lmp_node_polynom (node->data.operation.argv [1], var, exp, fact);
    default:
      break;
    }
  }

  res = lmp_node_polynom_fragment (node, var, &factel, &expel);
  lmp_stack_push (exp, expel);
  lmp_stack_push (fact, factel);

  return res;
}

lmp_node_t * lmp_node_create_polynom_fragment (lmp_node_t * var, lmp_node_t * fact, lmp_node_t * exp) {
  return lmp_node_new (OP, LMP_MULTIPLICATION, 2,
		       lmp_node_copy (fact),
		       lmp_node_new (OP, LMP_POWER, 2,
				     lmp_node_copy (var),
				     lmp_node_copy (exp)));
}

#define TOG(a,b) if (a && a->what == VAL && b && b->what == VAL) { \
  res = NEW_VAL (lmp_eval (lmp_node_new (OP, root->data.operation.op, 2, a, b))); \
  lmp_node_free (a); \
  lmp_node_free (b); \
  b = NULL; \
  a = res; \
}

#define CMB(a,b) if (b) { \
  a = lmp_node_new (OP, root->data.operation.op, 2, a, b); \
  b = NULL; \
}

/*
 *
 * (-a) + (-b) = - (a+b)
 * a + (-b)    = a-b
 * (-a) - (-b) = b-a
 * a - (-b)    = a+b
 * (-a) - b    = - (a+b)
 *
 * (-a) * (-b) = a*b
 * a * (-b)    = -(a*b)
 *
 * (-a) / (-b) = a/b
 * a / (-b)    = - (a/b)
 *
 * (-a)^2      = a^2   hereby check (real,even)
 * (-a)^3      = - a^3
 * 
 * sin (-a)    = - sin a
 * cos (-a)    = cos a
 * tan (-a)    = - tan a
 * cot (-a)    = - cot a
 *
 * - (-a)      = a
 */
lmp_node_t * lmp_tree_simplify_unary_minus (lmp_node_t * root) {
  lmp_node_t * pl, * pr;

  switch (root->what) {
  case OP:
    switch (root->data.operation.argc) {
    case 2:
      pr = lmp_tree_simplify_unary_minus (root->data.operation.argv [1]);
    case 1:
      pl = lmp_tree_simplify_unary_minus (root->data.operation.argv [0]);
    }
    switch (root->data.operation.op) {
    case LMP_ADDITION:
      if (pr->what == OP && pr->data.operation.op == LMP_UNARY_MINUS) {
	if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	  return lmp_node_new (OP, LMP_UNARY_MINUS, 2,
			       lmp_node_new (OP, LMP_ADDITION, 2,
					     pl->data.operation.argv [0],
					     pr->data.operation.argv [0]));
	}
	else {
	  return lmp_node_new (OP, LMP_SUBTRACTION, 2,
			       pl, pr->data.operation.argv [0]);
	}
      }
      else if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return lmp_node_new (OP, LMP_SUBTRACTION, 2,
			     pr, pl->data.operation.argv [0]);	
      }
      break;
    case LMP_SUBTRACTION:
      if (pr->what == OP && pr->data.operation.op == LMP_UNARY_MINUS) {
	if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	  return lmp_node_new (OP, LMP_SUBTRACTION, 2,
			       pr->data.operation.argv [0],
			       pl->data.operation.argv [0]);
	}
	else {
	  return lmp_node_new (OP, LMP_ADDITION, 2,
			       pl,
			       pr->data.operation.argv [0]);
	}
      }
      else if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return lmp_node_new (OP, LMP_UNARY_MINUS, 1,
			     lmp_node_new (OP, LMP_ADDITION, 2,
					   pl->data.operation.argv [0],
					   pr));
      }
      break;
    case LMP_MULTIPLICATION:
    case LMP_DIVISION:
      if (pr->what == OP && pr->data.operation.op == LMP_UNARY_MINUS) {
	if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	  return lmp_node_new (OP, root->data.operation.op, 2,
			       pl->data.operation.argv [0],
			       pr->data.operation.argv [0]);
	}
	else {
	  return lmp_node_new (OP, LMP_UNARY_MINUS, 1,
			       lmp_node_new (OP, root->data.operation.op, 2,
					     pl,
					     pr->data.operation.argv [0]));
	}
      }
      else if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return lmp_node_new (OP, LMP_UNARY_MINUS, 1,
			     lmp_node_new (OP, root->data.operation.op, 2,
					   pl->data.operation.argv [0],
					   pr));
      }
      break;
    case LMP_POWER:
      if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	if ((pr->what == VAL) && (MOD (pr->data.value, 2*ONE) == ZERO)) {
	  return lmp_node_new (OP, LMP_POWER, 2,
			       pl->data.operation.argv [0],
			       pr);
	}
	else {
	  return lmp_node_new (OP, LMP_UNARY_MINUS, 1,
			       lmp_node_new (OP, LMP_POWER, 2,
					     pl->data.operation.argv [0],
					     pr));
	}
      }
      break;
    case LMP_SINE:
    case LMP_TANGENT:
    case LMP_COTANGENT:
      if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return lmp_node_new (OP, LMP_UNARY_MINUS, 1,
			     lmp_node_new (OP, root->data.operation.op, 1,
					   pl->data.operation.argv [0]));
      }
      break;
    case LMP_COSINE:
      if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return lmp_node_new (OP, LMP_COSINE, 1, pl->data.operation.argv [0]);
      }
      break;
    case LMP_UNARY_MINUS:
      if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	return pl->data.operation.argv [0];
      }
      break;
    default:
    }
  default:
    return lmp_node_copy (root);
  }
}

/* More complex simplification:
 *
 * a*(b*x) = (a*b)*x where x is variable; a and b are values
 * ...
 */
lmp_node_t * lmp_tree_simplify_level2 (lmp_node_t * root) {
  lmp_node_t * pl, * pr;

  switch (root->what) {
  case CONTROL:
    {
      lmp_node_t * p;
      int i;

      pl = lmp_node_copy (root);
      for (i = 0; i < pl->data.control.argc; i++) {
	p = lmp_tree_simplify_level2 (pl->data.control.argv[i]);
	lmp_node_free (pl->data.control.argv [i]);
	pl->data.control.argv[i] = p;
      }

      return pl;
    }
    break;
  case OP:
    switch (root->data.operation.argc) {
    case 2:
      pr = lmp_tree_simplify_level2 (root->data.operation.argv [1]);
    case 1:
      pl = lmp_tree_simplify_level2 (root->data.operation.argv [0]);
    default:
    }    
    switch (root->data.operation.op) {
    case LMP_DIVISION:
      if (pr->what == OP && pr->data.operation.op == LMP_DIVISION) {
	lmp_node_t * p, * res;

	p = lmp_node_new (OP, LMP_DIVISION, 2,
			  lmp_node_new (OP, LMP_MULTIPLICATION, 2,
					pl, pr->data.operation.argv [1]),
			  pr->data.operation.argv [0]);
	res = lmp_tree_simplify_level1 (p);
	lmp_node_free (p);
	return res;
      }
      if (pl->what == OP && pl->data.operation.op == LMP_DIVISION) {
	lmp_node_t * p, * res;

	p = lmp_node_new (OP, LMP_DIVISION, 2,
			  pl->data.operation.argv [0],
			  lmp_node_new (OP, LMP_MULTIPLICATION, 2,
					pr, pl->data.operation.argv [1]));
	res = lmp_tree_simplify_level1 (p);
	lmp_node_free (p);
	return res;
      }
      if (pr->what == OP && pr->data.operation.op == LMP_POWER) {
	lmp_node_t * p, * res;

	p = lmp_node_new (OP, LMP_MULTIPLICATION, 2,
			  pl,
			  lmp_node_new (OP, LMP_POWER, 2,
					pr->data.operation.argv [0],
					lmp_node_new (OP, LMP_UNARY_MINUS, 1,
						      pr->data.operation.argv [1])));

	res = lmp_tree_simplify_full (p);
	lmp_node_free (p);

	return res;
      }
      {
	lmp_node_t * exp, * fact;
	if (pr->what == VAR && lmp_node_polynom_fragment (pl, pr, &fact, &exp)) {
	  lmp_node_t * res, * p;

	  p = lmp_node_new (OP, LMP_SUBTRACTION, 2,
			      exp,
			      NEW_VAL (ONE));

	  res = lmp_node_create_polynom_fragment (pr, fact, p);
	  lmp_node_free (p);
	  lmp_node_free (fact);
	  lmp_node_free (pr);

	  p = res;
	  res = lmp_tree_simplify_full (res);
	  lmp_node_free (p);

	  return res;
	}
      }
      break;
    case LMP_MULTIPLICATION:
      if ((pl->what == OP) &&
	  (pl->data.operation.op == LMP_EXP) &&
	  (pr->what == OP) &&
	  (pr->data.operation.op == LMP_EXP)) {
	lmp_node_t * res, * p;

	res = p = lmp_node_new (OP, LMP_EXP, 1,
				lmp_node_new (OP, LMP_ADDITION, 2,
					      pl->data.operation.argv [0],
					      pr->data.operation.argv [0]));
	res = lmp_tree_simplify_level1 (res);
	lmp_node_free (p);

	return res;
      }
      /* a*a^b   = a^(b+1) */
      if (pl->what == OP && pl->data.operation.op == LMP_POWER) {
	if (lmp_tree_equals (pl->data.operation.argv [0], pr)) {
	  lmp_node_t * res, * p;

	  p = lmp_node_new (OP, LMP_POWER, 2,
			    pr,
			    lmp_node_new (OP, LMP_ADDITION, 2,
					  pl->data.operation.argv [1],
					  NEW_VAL (ONE)));

	  lmp_node_free (pl->data.operation.argv [0]);

	  res = lmp_tree_simplify_level1 (p);
	  lmp_node_free (p);

	  return res;
	}
      }
      /* a^b*a   = a^(b+1) */
      if (pr->what == OP && pr->data.operation.op == LMP_POWER) {
	if (lmp_tree_equals (pr->data.operation.argv [0], pl)) {
	  lmp_node_t * res, * p;

	  p = lmp_node_new (OP, LMP_POWER, 2,
			    pl,
			    lmp_node_new (OP, LMP_ADDITION, 2,
					  pr->data.operation.argv [1],
					  NEW_VAL (ONE)));

	  lmp_node_free (pr->data.operation.argv [0]);

	  res = lmp_tree_simplify_level1 (p);
	  lmp_node_free (p);

	  return res;
	}
      }
      /* a^b*a^c = a^(b+c) */
      if (pl->what == OP && pl->data.operation.op == LMP_POWER &&
	  pr->what == OP && pr->data.operation.op == LMP_POWER) {
	if (lmp_tree_equals (pl->data.operation.argv [0], pr->data.operation.argv [0])) {
	  lmp_node_t * res, * p;

	  p = lmp_node_new (OP, LMP_POWER, 2,
			    pl->data.operation.argv [0],
			    lmp_node_new (OP, LMP_ADDITION, 2,
					  pl->data.operation.argv [1],
					  pr->data.operation.argv [1]));

	  lmp_node_free (pr->data.operation.argv [0]);

	  res = lmp_tree_simplify_level1 (p);
	  lmp_node_free (p);

	  return res;
	}
      }
    case LMP_ADDITION:
      if (root->data.operation.op == LMP_ADDITION) {
	if (pr->what == OP && pr->data.operation.op == LMP_UNARY_MINUS) {
	  lmp_node_t * res, * p;

	  res = p = lmp_tree_simplify_level1 (lmp_node_new (OP, LMP_SUBTRACTION, 2,
							    pl, pr->data.operation.argv [0]));
	  res = lmp_tree_simplify_level1 (res);
	  lmp_node_free (p);

	  return res;
	}
	if (pl->what == OP && pl->data.operation.op == LMP_UNARY_MINUS) {
	  lmp_node_t * res, * p;

	  res = p = lmp_tree_simplify_level1 (lmp_node_new (OP, LMP_SUBTRACTION, 2,
							    pr, pl->data.operation.argv [0]));
	  res = lmp_tree_simplify_level1 (res);
	  lmp_node_free (p);

	  return res;
	}
	if ((pl->what == OP) &&
	    (pl->data.operation.op == LMP_LOGARITHM) &&
	    (pr->what == OP) &&
	    (pr->data.operation.op == LMP_LOGARITHM)) {
	  lmp_node_t * res, * p;

	  res = p = lmp_node_new (OP, LMP_LOGARITHM, 1,
				  lmp_node_new (OP, LMP_MULTIPLICATION, 2,
						pl->data.operation.argv [0],
						pr->data.operation.argv [0]));
	  res = lmp_tree_simplify_level1 (res);
	  lmp_node_free (p);

	  return res;
	}
      }
      if (pl->what == VAL) {
	if ((pr->what == OP) && (pr->data.operation.op == root->data.operation.op)) {
	  if (pr->data.operation.argv[0]->what == VAL) {
	    lmp_node_t * tmp1, * tmp2, * res;

	    tmp1 = lmp_node_copy (pr->data.operation.argv[0]);
	    tmp2 = lmp_node_copy (pr->data.operation.argv[1]);
	    lmp_node_free (pr);
	    res = lmp_tree_simplify_level1 
	      (lmp_node_new
	       (OP, root->data.operation.op, 2,
		NEW_VAL (lmp_eval 
			 (lmp_node_new 
			  (OP, root->data.operation.op, 2, pl, tmp1))),
		tmp2));

	    lmp_node_free (tmp1);
	    lmp_node_free (tmp2);
	    lmp_node_free (pl);

	    return res;
	  }
	  else if (pr->data.operation.argv[1]->what == VAL) {
	    lmp_node_t * tmp1, * tmp2, * res;

	    tmp1 = lmp_node_copy (pr->data.operation.argv[0]);
	    tmp2 = lmp_node_copy (pr->data.operation.argv[1]);
	    lmp_node_free (pr);
	    res = lmp_tree_simplify_level1
	      (lmp_node_new
	       (OP, root->data.operation.op, 2,
		NEW_VAL (lmp_eval 
			 (lmp_node_new 
			  (OP, root->data.operation.op, 2, pl, tmp2))),
		tmp1));

	    lmp_node_free (tmp1);
	    lmp_node_free (tmp2);
	    lmp_node_free (pl);

	    return res;
	  }
	}
      }
      if (pr->what == VAL) {
	if ((pl->what == OP) && (pl->data.operation.op == root->data.operation.op)) {
	  if (pl->data.operation.argv[0]->what == VAL) {
	    lmp_node_t * tmp1, * tmp2, * res;

	    tmp1 = lmp_node_copy (pl->data.operation.argv[0]);
	    tmp2 = lmp_node_copy (pl->data.operation.argv[1]);
	    lmp_node_free (pl);
	    res = lmp_tree_simplify_level1
	      (lmp_node_new
	       (OP, root->data.operation.op, 2,
		NEW_VAL (lmp_eval
			 (lmp_node_new
			  (OP, root->data.operation.op, 2, pr, tmp1))),
		tmp2));
	    lmp_node_free (tmp1);
	    lmp_node_free (tmp2);
	    lmp_node_free (pr);

	    return res;
	  }
	  else if (pl->data.operation.argv[1]->what == VAL) {
	    lmp_node_t * tmp1, * tmp2, * res;

	    tmp1 = lmp_node_copy (pl->data.operation.argv[0]);
	    tmp2 = lmp_node_copy (pl->data.operation.argv[1]);
	    lmp_node_free (pl);
	    res = lmp_tree_simplify_level1
	      (lmp_node_new
	       (OP, root->data.operation.op, 2,
		NEW_VAL (lmp_eval
			      (lmp_node_new
			       (OP, root->data.operation.op, 2, pr, tmp2))),
		tmp1));

	    lmp_node_free (tmp1);
	    lmp_node_free (tmp2);
	    lmp_node_free (pr);

	    return res;	    
	  }
	}
      }
      if ((pl->what == OP) &&
	  (pl->data.operation.op == root->data.operation.op) &&
	  (pr->what == OP) &&
	  (pr->data.operation.op == root->data.operation.op)) {
	lmp_node_t * tmp1, * tmp2, * tmp3, * tmp4, * res;

	tmp1 = lmp_node_copy (pl->data.operation.argv [0]);
	tmp2 = lmp_node_copy (pl->data.operation.argv [1]);
	tmp3 = lmp_node_copy (pr->data.operation.argv [0]);
	tmp4 = lmp_node_copy (pr->data.operation.argv [1]);

	TOG (tmp1,tmp2);
	TOG (tmp1,tmp3);
	TOG (tmp1,tmp4);
	TOG (tmp2,tmp3);
	TOG (tmp2,tmp4);
	TOG (tmp3,tmp4);

	CMB (tmp1,tmp2);
	CMB (tmp1,tmp3);
	CMB (tmp1,tmp4);

	res = tmp1;
	return res;
      }
      // NO break HERE: following code needs to be executed
    default:
    }

    switch (root->data.operation.argc) {
    case 1:
      return lmp_tree_simplify_level1 (lmp_node_new (OP, root->data.operation.op, 1, pl));
    case 2:
      return lmp_tree_simplify_level1 (lmp_node_new (OP, root->data.operation.op, 2, pl, pr));
    }
    break;
  default:
  }

  return lmp_node_copy (root);
}

/* This is strait forward simplification.
 *
 * 1*a = a
 * 0*a = 0
 * 0+a = a
 * ...
 *
 * exp(ln(x)) = ln(exp(x)) = x
 */
lmp_node_t * lmp_tree_simplify_level1 (lmp_node_t * root) {
  lmp_node_t * p, * pl, * pr;

  p = root;
  if (p) {
    if (lmp_tree_evaluable (p))
      return NEW_VAL (lmp_eval (p));

    switch (p->what) {
    case OP:
      switch (root->data.operation.argc) {
      case 2:
	pr = lmp_tree_simplify_level1 (root->data.operation.argv [1]);
      case 1:
	pl = lmp_tree_simplify_level1 (root->data.operation.argv [0]);
      default:
      }
      switch (root->data.operation.op) {
      case LMP_MULTIPLICATION:
	if (lmp_tree_equals (pl, pr)) {
	  lmp_node_free (pr);
	  return lmp_node_new (OP, LMP_POWER, 2,
			       pl,
			       NEW_VAL (2*ONE));
	}

	if (pl->what == VAL) {
	  if (pl->data.value == ZERO) {
	    lmp_node_free (pr);
	    return pl;
	  }
	  else if (pl->data.value == ONE) {
	    lmp_node_free (pl);
	    return pr;
	  }
	}

	if (pr->what == VAL) {
	  if (pr->data.value == ZERO) {
	    lmp_node_free (pl);
	    return pr;
	  }
	  else if (pr->data.value == ONE) {
	    lmp_node_free (pr);
	    return pl;
	  }
	  else if (pl->what != VAL)
	    return lmp_node_new (OP, root->data.operation.op, 2, pr, pl);
	}

	return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);

      case LMP_DIVISION:
	if (lmp_tree_equals (pl, pr)) {
	  lmp_node_free (pr);
	  lmp_node_free (pl);
	  return NEW_VAL (ONE);
	}
	if (pr->what == VAL) {
	  if (pr->data.value == ONE) {
	    lmp_node_free (pr);
	    return pl;
	  }
	  if (pr->data.value == -1*ONE) {
	    lmp_node_free (pr);
	    return lmp_node_new (OP, LMP_UNARY_MINUS, 1, pl);
	  }
	}

	return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);

      case LMP_SUBTRACTION:
	if (lmp_tree_equals (pl, pr)) {
	  lmp_node_free (pl);
	  lmp_node_free (pr);
	  return NEW_VAL (ZERO);
	}
	if (pl->what == VAL) {
	  if (pl->data.value == ZERO) {
	    lmp_node_free (pl);
	    return lmp_node_new (OP, LMP_UNARY_MINUS, 1, pr);
	  }
	}
	if (pr->what == VAL) {
	  if (pr->data.value == ZERO) {
	    lmp_node_free (pr);
	    return lmp_node_new (OP, LMP_UNARY_MINUS, 1, pl);
	  }
	}
	return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);
      case LMP_ADDITION:
	if (lmp_tree_equals (pl, pr)) {
	  lmp_node_free (pr);
	  return lmp_node_new (OP, LMP_MULTIPLICATION, 2,
			       NEW_VAL (2*ONE),
			       pl);
	}
	if (pl->what == VAL) {
	  if (pl->data.value == ZERO) {
	    lmp_node_free (pl);
	    return pr;
	  }
	}
	if (pr->what == VAL) {
	  if (pr->data.value == ZERO) {
	    lmp_node_free (pr);
	    return pl;
	  }
	}

	return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);
      case LMP_POWER:
	if (pl->what == OP && pl->data.operation.op == LMP_POWER) {
	  lmp_node_t * res;
	  res = lmp_tree_simplify_level1
	    (lmp_node_new (OP, LMP_POWER, 2,
			   pl->data.operation.argv[0],
			   lmp_node_new (OP, LMP_MULTIPLICATION, 2,
					 pr, pl->data.operation.argv [1])));
	  lmp_node_free (pl);
	  lmp_node_free (pr);

	  return res;
	}
	if (pl->what == VAL) {
	  if (pl->data.value == ZERO) {
	    lmp_node_free (pr);
	    return pl;
	  }
	  else if (pl->data.value == ONE) {
	    lmp_node_free (pr);
	    return pl;
	  }
	}
	if (pr->what == VAL) {
	  if (pr->data.value == ZERO) {
	    lmp_node_free (pr);
	    lmp_node_free (pl);
	    return NEW_VAL (ONE);
	  }
	  else if (pr->data.value == ONE) {
	    lmp_node_free (pr);
	    return pl;
	  }
	}

	return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);
      case LMP_EXP:
	if (pl->what == OP && pl->data.operation.op == LMP_LOGARITHM) {
	  return pl->data.operation.argv [0];
	}

	return lmp_node_new (OP, root->data.operation.op, 1, pl);

      case LMP_LOGARITHM:
	if (pl->what == OP && pl->data.operation.op == LMP_EXP) {
	  return pl->data.operation.argv [0];
	}

	return lmp_node_new (OP, root->data.operation.op, 1, pl);

      default:
	switch (root->data.operation.argc) {
	case 2:
	  return lmp_node_new (OP, root->data.operation.op, 2, pl, pr);
	case 1:
	  return lmp_node_new (OP, root->data.operation.op, 1, pl);
	}
      }
    case CONTROL:
      {
	lmp_node_t * p;
	int i;

	pl = lmp_node_copy (root);
	for (i = 0; i < pl->data.control.argc; i++) {
	  p = lmp_tree_simplify_level1 (pl->data.control.argv[i]);
	  lmp_node_free (pl->data.control.argv [i]);
	  pl->data.control.argv[i] = p;
	}

	return pl;
      }
    default:
      return lmp_node_copy (root);
    }
  }

  return lmp_node_copy (root);
}

int lmp_tree_equals (lmp_node_t * a, lmp_node_t * b) {
  lmp_node_t * tmp;
  int res;

  switch (a->what) {
  case OP:
    if (lmp_tree_evaluable (a)) { // it's just a value
      if (lmp_tree_evaluable (b)) { // could be equal only if b is a value
	return lmp_eval (a) == lmp_eval (b);
      }
      else
	return 0;
    }
    else {
      if ((b->what == OP) && (b->data.operation.op == a->data.operation.op)) {
	int i;

	for (i = 0; i < a->data.operation.argc; i++) {
	  if (!lmp_tree_equals (a->data.operation.argv [i], b->data.operation.argv [i]))
	    return 0;
	}

	return 1;
      }
      else
	return 0; // THIS NEEDS COMPLETATION
    }
    break;
  case VAR:
    if (lmp_tree_contains (b, a)) { // could be equal only if b contains a
      tmp = lmp_tree_simplify_level1 (b);
      if ((b->what == VAR) && (!strcmp (b->data.variable, a->data.variable))) {
	lmp_node_free (tmp);
	return 1;
      }
      else {
	lmp_node_free (tmp);
	return 0;
      }
    }
    else
      return 0;
    break;
  case VAL:
    if (lmp_tree_evaluable (b)) { // could be equal only if b is a value
      return lmp_eval (b) == a->data.value;
    }
    else
      return 0;
    break;
  case CONTROL:
    tmp = lmp_proc_control (a);
    res = lmp_tree_equals (tmp, b);
    lmp_node_free (tmp);
    break;
  }

  return res;
}

lmp_node_t * lmp_tree_var_eval (lmp_node_t * var, lmp_node_t * root) {
  lmp_node_t * res;
  int i;

  if (root) {
    i = 0;
    switch (root->what) {
    case OP:
      res = lmp_node_new (OP, root->data.operation.op, root->data.operation.argc, NULL, NULL);
      while (i < root->data.operation.argc) {
	res->data.operation.argv [i] = lmp_tree_var_eval (var, root->data.operation.argv [i]);
	i++;
      }
      break;
    case CONTROL:
      res = lmp_node_new (CONTROL, root->data.control.op, root->data.control.argc, NULL, NULL, NULL, NULL);
      while (i < root->data.control.argc) {
	res->data.control.argv [i] = lmp_tree_var_eval (var, root->data.control.argv [i]);
	i++;
      }
      break;
    case VAL:
      res = lmp_node_copy (root);
      break;
    case VAR:
      if (!strcmp (root->data.variable, var->data.variable)) {
	lmp_value_type v;

	lmp_vt_lookup (root->data.variable, &v);
	res = NEW_VAL (v);
      }
      else
	res = lmp_node_copy (root);
      break;
    }
  }

  return res;
}

int lmp_tree_contains (lmp_node_t * tree, lmp_node_t * var) {
  lmp_node_t * p;
  int i;

  if (var->what != VAR) {
    lmp_error ("INTERNAL ERROR: call `lmp_tree_contains' with no variable but %s.\n", lmp_infix_from_tree (var, 0));
    return 0;
  }

  p = tree;
  if (p) {
    i = 0;
    switch (p->what) {
    case OP:
      while (i < p->data.operation.argc)
	if (lmp_tree_contains (p->data.operation.argv [i++], var))
	  return 1;
      break;
    case CONTROL:
      while (i < p->data.control.argc)
	if (lmp_tree_contains (p->data.control.argv [i++], var))
	  return 1;
      break;
    case VAL:
      break;
    case VAR:
      if (!strcmp (p->data.variable, var->data.variable))
	return 1;
      break;
    }
  }

  return 0;
}

int lmp_tree_evaluable (lmp_node_t * tree) {
  lmp_node_t * p;
  int i;

  p = tree;
  if (p) {
    i = 0;
    switch (p->what) {
    case OP:
      while (i < p->data.operation.argc)
	if (!lmp_tree_evaluable (p->data.operation.argv [i++]))
	  return 0;
      break;
    case CONTROL:
      while (i < p->data.control.argc)
	if (!lmp_tree_evaluable (p->data.control.argv [i++]))
	  return 0;
      break;
    case VAL:
      break;
    case VAR:
	return 0;
      break;
    }
  }

  return 1;
}

int lmp_tree_balance_build_stack (lmp_stack_t * stack, lmp_node_t * p, lmp_operator op) {
  if (p->what == OP && p->data.operation.op == op) {
    lmp_tree_balance_build_stack (stack, p->data.operation.argv [0], op);
    lmp_tree_balance_build_stack (stack, p->data.operation.argv [1], op);

    return 1;
  }
  if (p->what == OP) {
    if (op == LMP_ADDITION) {
      if (p->data.operation.op == LMP_UNARY_MINUS && p->data.operation.argv[0]->what == OP) {
	if (p->data.operation.argv[0]->data.operation.op == LMP_ADDITION) {
	  lmp_tree_balance_build_stack (stack, lmp_node_new
					(OP, LMP_UNARY_MINUS, 1,
					 p->data.operation.argv[0]->data.operation.argv [0]), op);
	  lmp_tree_balance_build_stack (stack, lmp_node_new
					(OP, LMP_UNARY_MINUS, 1,
					 p->data.operation.argv[0]->data.operation.argv [1]), op);

	  return 1;
	}
	else if (p->data.operation.argv[0]->data.operation.op == LMP_UNARY_MINUS) {
	  lmp_tree_balance_build_stack (stack, p->data.operation.argv[0]->data.operation.argv[0], op);

	  return 1;
	}
	else if (p->data.operation.argv[0]->data.operation.op == LMP_SUBTRACTION) {
	  lmp_tree_balance_build_stack (stack, p->data.operation.argv[0]->data.operation.argv[1], op);
	  lmp_tree_balance_build_stack (stack, lmp_node_new
				      (OP, LMP_UNARY_MINUS, 1,
				       p->data.operation.argv[0]->data.operation.argv[0]), op);
	  return 1;
	}
      }
      else if (p->data.operation.op == LMP_SUBTRACTION) {
	lmp_tree_balance_build_stack (stack, p->data.operation.argv [0], op);
	lmp_tree_balance_build_stack (stack, lmp_node_new
				      (OP, LMP_UNARY_MINUS, 1, p->data.operation.argv [1]), op);
	return 1;
      }
    }
  }

  lmp_stack_push (stack, lmp_tree_balance (p));

  return 1;
}

lmp_node_t * lmp_tree_balance_build (lmp_stack_t * stack, lmp_operator op) {
  lmp_stack_t * r, * l;

  if (stack->si > 2) { // split if more than 2 elements on stack
    lmp_stack_split (stack, &l, &r);
    return lmp_node_new (OP, op, 2,
			 lmp_tree_balance_build (l, op),
			 lmp_tree_balance_build (r, op));
  }
  else if (stack->si == 2) {
    return lmp_node_new (OP, op, 2, lmp_node_copy (lmp_stack_get (stack, 0)), lmp_node_copy (lmp_stack_get (stack, 1)));
  }
  else if (stack->si == 1) {
    return lmp_node_copy (lmp_stack_get (stack, 0));
  }
  else if (stack->si == 0) {
    if (op = LMP_ADDITION)
      return NEW_VAL (ZERO);
    else
      return lmp_node_copy (lmp_stack_get (stack, 0));
  }
  else {
    lmp_error ("Stack underrun in lmp_tree_balance_build: %d!\n", stack->si);
    return NULL;
  }
}

lmp_stack_t * lmp_tree_balance_reduce_stack (lmp_stack_t * stack, lmp_operator op) {
  lmp_stack_t * res;
  lmp_node_t * r, * l;
  int i, j;
  int * v;

  res = lmp_stack_new ();

  /*
    0,1..n: (0,1)(0,2)(0,3) ... (0,n)
    1,2..n:      (1,2)(1,3) ... (1,n)
    ...
    n-1,n:                    (n-1,n)
   */

  v = malloc (sizeof (int) * stack->si);
  for (i = 0; i < stack->si; i++)
    v [i] = 1;

  for (i = 0; i < stack->si; i++) {
    for (j = i+1; j < stack->si; j++) {
      l = lmp_stack_get (stack, i);
      r = lmp_stack_get (stack, j);
      if (lmp_tree_equals (l, r)) {
	if (v [j]) {
	  v [j]--;
	  v [i]++;
	}
      }
      else if (lmp_tree_equals (lmp_node_new (OP, LMP_UNARY_MINUS, 1, l), r)) {
	if (v [j] && v [i]) {
	  v [j]--;
	  v [i]--;
	}
      }
      else if (lmp_tree_equals (lmp_node_new (OP, LMP_UNARY_MINUS, 1, r), l)) {
	if (v [j] && v [i]) {
	  v [j]--;
	  v [i]--;
	}
      }
    }
  }

  for (i = 0; i < stack->si; i++) {
    switch (op) {
    case LMP_MULTIPLICATION:
      if (v [i] > 1)
	lmp_stack_push (res, lmp_node_new (OP, LMP_POWER, 2, 
					   lmp_stack_get (stack, i),
					   NEW_VAL (v[i]*ONE)));
      else if (v [i] == 1)
	lmp_stack_push (res, lmp_stack_get (stack, i));
      break;
    case LMP_ADDITION:
      if (v [i] > 1)
	lmp_stack_push (res, lmp_node_new (OP, LMP_MULTIPLICATION, 2, 
					   NEW_VAL (v[i]*ONE),
					   lmp_stack_get (stack, i)));
      else if (v [i] == 1)
	lmp_stack_push (res, lmp_stack_get (stack, i));
      break;
    default:
    }
  }

  free (v);
  return res;
}

lmp_node_t * lmp_tree_balance (lmp_node_t * root) {
  auto lmp_node_t * l, * r, * p;
  auto lmp_stack_t * stack;

  switch (root->what) {
  case OP:
    switch (root->data.operation.argc) {
      // Recursion is too much work, however ...
    case 2:

#ifdef REC_BAL
      r = lmp_tree_balance (root->data.operation.argv [1]);
#else
      r = root->data.operation.argv [1];
#endif
    case 1:
#ifdef REC_BAL
      l = lmp_tree_balance (root->data.operation.argv [0]);
#else
      l = root->data.operation.argv [0];
#endif
    }

    switch (root->data.operation.op) {
    case LMP_SUBTRACTION:
      stack = lmp_stack_new ();

      lmp_tree_balance_build_stack (stack, l, LMP_ADDITION);
      lmp_tree_balance_build_stack (stack, lmp_node_new (OP, LMP_UNARY_MINUS, 1, r), LMP_ADDITION);
      
      if (stack->si > 0) { // only rebuild if there is something to be rebuild
	lmp_stack_t * news;

	news = lmp_tree_balance_reduce_stack (stack, LMP_ADDITION);
	lmp_stack_free (stack);
	stack = news;
	p = lmp_tree_balance_build (stack, LMP_ADDITION);
	lmp_stack_free (stack);
	return p;
      }
      break;
    case LMP_MULTIPLICATION:
    case LMP_ADDITION:
      stack = lmp_stack_new ();

      lmp_tree_balance_build_stack (stack, l, root->data.operation.op);
      lmp_tree_balance_build_stack (stack, r, root->data.operation.op);

      if (stack->si > 2) { // only rebuild if there is something to be rebuild
	lmp_stack_t * news;

	news = lmp_tree_balance_reduce_stack (stack, root->data.operation.op);
	lmp_stack_free (stack);
	stack = news;
	p = lmp_tree_balance_build (stack, root->data.operation.op);
	lmp_stack_free (stack);
	return p;
      }
    default:
      //      lmp_node_free (r);
      //      lmp_node_free (l);
      switch (root->data.operation.argc) {
      case 1:
	return lmp_node_new (OP, root->data.operation.op, 1, lmp_tree_balance (l));
      case 2:
	return lmp_node_new (OP, root->data.operation.op, 2, lmp_tree_balance (l), lmp_tree_balance (r));
      }
    }
    break;
  default:
    return lmp_node_copy (root);
  }

  return lmp_node_copy (root);
}

lmp_node_t * lmp_tree_simplify (lmp_node_t * root, lmp_simpl_t simpl) {
  lmp_node_t * p, * pold;

  pold = p = lmp_node_copy (root);

  if (simpl.do_balance)
    pold = p = lmp_tree_balance (p);

  if (simpl.do_level1) {
    p = lmp_tree_simplify_level1 (p);
    lmp_node_free (pold);
    pold = p;
  }

  if (simpl.do_level2) {
    p = lmp_tree_simplify_level2 (p);
    lmp_node_free (pold);
    pold = p;
  }

  if (simpl.do_unary_minus) {
    p = lmp_tree_simplify_unary_minus (p);
    lmp_node_free (pold);
    pold = p;
  }

  return p;
}

lmp_node_t * lmp_tree_simplify_full (lmp_node_t * root) {
  lmp_simpl_t simpl;

  simpl.do_balance = 1;
  simpl.do_level1  = 1;
  simpl.do_level2  = 1;
  simpl.do_unary_minus = 1;

  return lmp_tree_simplify (root, simpl);
}
